%==========================================================
% ecf_update.m - Update filter estimate with new IMU 
%                measurements.
%
% x - filter state structure
% gyr - gyro angular velocity measurement
% k_acc - tuning parameter deciding impact of accelerometer 
%         measurement
% acc - accelerometer measurement in body coordinates.
% acc_ref - acceleration reference vector in NED.
% k_mag - tuning parameter deciding impact of magnetometer
%         measurement
% mag - magnetometer measurement in body coordinates.
% mag_ref - magnetic field reference vector in NED.
% dt - time since last update.
%==========================================================
function [x] = ecf_update(x, gyr, k_acc, acc, acc_ref, ...
                                 k_mag, mag, mag_ref, dt)

%Rotate reference vectors to body coordinates using the
%current estimate of the attitude:
acc_ref = x.R*acc_ref';
mag_ref = x.R*mag_ref';

%Normalize vectors:
acc = acc / norm(acc);
acc_ref = acc_ref / norm(acc_ref);
mag = mag / norm(mag);
mag_ref = mag_ref / norm(mag_ref);

%Construct w_mes:
w_mes = 0.5*(k_acc*cross(acc, acc_ref) + ...
                             k_mag*cross(mag, mag_ref));

%Calculate derivatives:
q_hat_dot = T_q(x.q_hat)*(gyr - x.gyr_bias_hat + ...
                                          x.k_p*w_mes)';
gyr_bias_hat_dot = -x.k_i*w_mes;

%Integrate states using forward euler:
x.q_hat = x.q_hat + dt*q_hat_dot';
x.gyr_bias_hat = x.gyr_bias_hat + dt*gyr_bias_hat_dot;

%Normalize quaternion:
x.q_hat = x.q_hat / norm(x.q_hat);

%Calculate rotation matrix:
x.R = q2rot(x.q_hat);