%filtering.m -  Filtering sensor data using Mahoney (ECF) and Hua 
%               observers.

% f - structure with sensor values and reference vectors used:
clear f;
f.t_imu = sensor_a.time'; %IMU sample timestamps.
f.acc = sensor_a.signals.values; %Specific force measured in BODY.
f.gyr = sensor_omega.signals.values; %Angular velocity measured in BODY.
f.mag = sensor_m.signals.values; %Magnetic field measured in BODY.
f.t_gps = sensor_p.time'; %GPS sample timestamps.
f.p = sensor_p.signals.values; %Position from GPS
f.p_dot = sensor_p_dot.signals.values; %Velocity from GPS.
f.g_ref = [0 0 -9.81]; %Reference vector if not using feedback from Hua.
f.m_ref = [0.9548 0.0406 0.2946]; %Magnetic field vector (normalized)

%Filter gains:
k_i = 0.05;
k_p = 1.9;
k_a = 10;
k_m = 5;
%k_1 = 3;
k_1 = 0.5;
%k_v = 0.1;
k_v = 0.001;

%Initialize filter states (kp, ki, euler_init, gyr_bias_init)
ecf_state = ecf_init(k_p, k_i, [0 0 0], [0 0 0]);
ecf_state2 = ecf_init(k_p, k_i, [0 0 0], [0 0 0]);

%Initialize hua filter states:
hua_state = hua_init([20 0 0]'); 

%Filter data:
i=1;
t_prev = 0;
for t=f.t_imu
    
    %Find dt: 
    dt = t - t_prev;
    
    %Predict HUA filter: (x, w, a, dt)
    hua_state = hua_predict(hua_state, ...
                            f.gyr(i, :)', ...
                            f.acc(i, :)', ...
                            dt);
    
    %Update hua filter with gps measurement:
    if (mod(i, 10) == 0)
       %Update HUA filter: (x, k_1, k_v, vel, a, dt)
       hua_state = hua_update(  hua_state, ...
                                k_1, k_v, ...
                                f.p_dot(i/10, :)', ...
                                f.acc(i,:)', ...
                                10*dt);
    end
    
    
    %Specific force reference vector to be used in ECF:
    f_ref = hua_state.v_hat_dot' - [0 0 9.81];
    %f_ref = (hua_state.Q_hat*f.acc(i,:)')';
    f_ref_g = (ecf_state.R'*f.acc(i,:)')';
    
    f_diff(i) = norm(f_ref - f_ref_g);
    
    %Do not use specific force vector measurement if it is too large:
    %k_a_scaling = 
    
    %Prediction and update of ECF attitude estimation:
    ecf_state = ecf_update( ecf_state, f.gyr(i, :), ...
                            k_a, f.acc(i, :), f_ref, ...
                            k_m, f.mag(i,:), f.m_ref, dt);
                        
    ecf_state2 = ecf_update(ecf_state2, f.gyr(i, :), ...
                            k_a, f.acc(i, :), f.g_ref, ... 
                            k_m, f.mag(i,:), f.m_ref, dt);
        
    %Save values for plotting:
    f.q_hat(i, :) = ecf_state.q_hat;
    f.v_body_hat(i, :) = ecf_state.R*hua_state.v_hat;
    filter_euler(i, :) = rot2e(ecf_state.R);
    filter_euler_g(i, :) = rot2e(ecf_state2.R);
    gyr_bias(i, :) = ecf_state.gyr_bias_hat;
    gyr_bias_g(i, :) = ecf_state2.gyr_bias_hat;
    hua_v_hat(i, :) = hua_state.v_hat;
    hua_an_hat(i, :) = hua_state.v_hat_dot;
    
    i = i + 1;
    t_prev = t;
end

%Store results:
f.euler_hat = filter_euler;
f.b_hat = gyr_bias;
f.v_hat = hua_v_hat;
f.f_hat = hua_an_hat;

%Calculate estimated airspeed: 
f.norm_v_hat = sqrt(f.v_body_hat(:, 1).^2 + ...
                    f.v_body_hat(:, 2).^2 + ...
                    f.v_body_hat(:, 3).^2);

%Calculate estimated dynamic pressure:
f.q_bar_hat = 0.5*1.2*f.norm_v_hat.^2;

%Calculate estimated angle of attack:
f.alpha_hat = atan(f.v_body_hat(:, 3) ./ f.v_body_hat(:, 1));

%Calculate estimated side-slip angle:
f.beta_hat = asin(f.v_body_hat(:, 2) ./ f.norm_v_hat);

%========================================================================
% Visualization of filtered results:
%========================================================================


%Plot estimated attitude:
figure(1);
clf;
subplot(3, 1, 1);
h = plot(sim_euler.time, sim_euler.signals.values(:, 1)*180/pi, 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(sensor_a.time, [filter_euler(:, 1)]*180/pi, 'k--');
set(h, 'LineWidth', 1);
h = plot(sensor_a.time, [filter_euler_g(:, 1)]*180/pi, 'r--');
set(h, 'LineWidth', 1);
grid on;
title('Estimated attitude', 'FontSize', 14);
legend('True', 'Est. with corrections', 'Est. without corrections', 'location', 'NorthEast', 'Orientation', 'vertical');
ylabel('Roll (deg)');

subplot(3, 1, 2);
h = plot(sim_euler.time, sim_euler.signals.values(:, 2)*180/pi, 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(sensor_a.time, [filter_euler(:, 2)]*180/pi, 'k--');
set(h, 'LineWidth', 1);
h = plot(sensor_a.time, [filter_euler_g(:, 2)]*180/pi, 'r--');
set(h, 'LineWidth', 1);
grid on;
ylabel('Pitch (deg)');

subplot(3, 1, 3);
h = plot(sim_euler.time, sim_euler.signals.values(:, 3)*180/pi, 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(sensor_a.time, [filter_euler(:, 3)]*180/pi, 'k--');
set(h, 'LineWidth', 1);
h = plot(sensor_a.time, [filter_euler_g(:, 3)]*180/pi, 'r--');
set(h, 'LineWidth', 1);
grid on;
ylabel('Yaw (deg)');
xlabel('Time (s)');

print('-dpdf', '../report/gfx/filtered_att.pdf');

%Estimated bias plot:
figure(2);
clf;
h = plot(sensor_a.time, gyr_bias);
set(h, 'LineWidth', 2);
hold on;
h = plot(sensor_a.time, gyr_bias_g, '--');
title('Estimated gyro bias with and without acceleration corrections', 'FontSize', 14);
ylabel('Estimated bias (deg/s)');
xlabel('Time');
%ylim([-0.05 0.05]);
print('-dpdf', '../report/gfx/est_bias.pdf');


%Estimated specific force:
figure(3);
clf;
h = plot(p_ddot.time, p_ddot.signals.values);
title('Estimated specific force using Hua observer');
set(h, 'LineWidth', 2);
hold on;
h = plot(sensor_a.time, hua_an_hat, '.');
set(h, 'LineWidth', 1);
xlabel('Time (s)');
ylabel('Spesific force (m/s^2)');
legend('True f_x', 'True f_y', 'True f_z', 'Est. f_x', 'Est. f_y', 'Est. f_z');

print('-dpdf', '../report/gfx/est_spec_force.pdf');


%NED velocity:
figure(4);
clf;
h = plot(sim_p_dot.time, sim_p_dot.signals.values(:, 1), 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(sim_p_dot.time, sim_p_dot.signals.values(:, 2), 'c');
set(h, 'LineWidth', 2);
set(get(get(h,'Annotation'),'LegendInformation'),'IconDisplayStyle','off'); 
h = plot(sim_p_dot.time, sim_p_dot.signals.values(:, 3), 'c');
set(h, 'LineWidth', 2);
set(get(get(h,'Annotation'),'LegendInformation'),'IconDisplayStyle','off'); 
plot(sensor_a.time, hua_v_hat, '--');
title('Estimated velocity expressed in NED coordinates', 'FontSize', 14);
legend('True value', 'North', 'East', 'Down');
xlabel('Time (s)');
ylabel('Velocity (m/s)');
grid on;
print('-dpdf', '../report/gfx/est_velocity_ned.pdf');

%BODY velocity:
figure(7);
clf;
subplot(3, 1, 1);
h = plot(sim_p_dot.time, sim_v_body.signals.values(:, 1), 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(f.t_imu, f.v_body_hat(:, 1), '--');
set(h, 'LineWidth', 1);
title('Estimated velocities expressed in BODY coordinates', 'FontSize', 14);
legend('True velocity', 'Estimated');
ylabel('u (m/s)');
subplot(3, 1, 2);
h = plot(sim_p_dot.time, sim_v_body.signals.values(:, 2), 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(f.t_imu, f.v_body_hat(:, 2), '--');
set(h, 'LineWidth', 1);
ylabel('v (m/s)');
subplot(3, 1, 3);
h = plot(sim_p_dot.time, sim_v_body.signals.values(:, 3), 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(f.t_imu, f.v_body_hat(:, 3), '--');
set(h, 'LineWidth', 1);
xlabel('Time (s)');
ylabel('w (m/s)');
grid on;

print('-dpdf', '../report/gfx/est_velocity_body.pdf');


%Wind angles:
figure(5);
clf;
subplot(2, 1, 1);
h = plot(alpha.time, alpha.signals.values*180.0/pi, 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(f.t_imu, f.alpha_hat*180.0/pi, 'k--');
set(h, 'LineWidth', 1);
ylabel('\alpha (deg)');
grid on;
title('Estimated and actual wind angles', 'FontSize', 14);
legend('True value', 'Estimated', 'location', 'NorthEast', 'Orientation', 'horizontal');

subplot(2, 1, 2);
h = plot(beta.time, beta.signals.values*180.0/pi, 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(f.t_imu, f.beta_hat*180.0/pi, 'k--');
set(h, 'LineWidth', 1);
ylabel('\beta (deg)');
grid on;
xlabel('Time (s)');

print('-dpdf', '../report/gfx/est_wind_angles.pdf');

%Dynamic pressure:
figure(6);
clf;
h = plot(sim_q_bar.time, sim_q_bar.signals.values, 'c');
set(h, 'LineWidth', 2);
hold on;
h = plot(f.t_imu, f.q_bar_hat, 'k--');
set(h, 'LineWidth', 1);
ylabel('Dynamic pressure (N/m^2)');
grid on;
title('Estimated and actual dynamic pressure', 'FontSize', 14);
legend('True value', 'Estimated', 'location', 'NorthEast', 'Orientation', 'horizontal');
grid on;
xlabel('Time (s)');

print('-dpdf', '../report/gfx/est_dyn_pres.pdf');