%mpoly - moving polynomial smoothing, derivation, interpolation
%         and extrapolation. a
% Author: Thomas Ingebretsen
%       + y[N]   - input data points
%       + t[N]   - input timestamps (lenght(t) == length(y))
%       + m      - polynomial order
%       + nl     - number of points to be included in fit to the left.
%       + nr     - number of points to be included in fit to the right.
%       + ld     - number of times to differentiate the function.
%       + t_grid - time-grid of the output function.
function [ret_y, ret_t] = mpoly(y, t, m, nl, nr, ld, t_grid)

%Number of input data points (N):
N = length(y);
%Number of output data points (N_OUT):
N_OUT = length(t_grid);

%Check input parameters:
if (nl+nr+1 < m)
    disp('mpoly: error: polynomial degree m should be >= window width...');
    return;
end
if (N ~= length(t))
    disp('mpoly: error: size of y and t must agree,,,');
    return;
end
if (ld > m) 
    disp('mpoly: error: ld > m');
    return;
end

%Preallocate:
ret_y = zeros(N_OUT, 1);

%Give polonomial numbers to all points in the output time grid (t_grid):
pnum = zeros(1, N_OUT);
j = 1;
for i=1:N
    %Next i to check:
    if (i==N)
        next_i = i;
    else 
        next_i = i+1;
    end
    
    %Assign polynomials to output points based on length from polynomial
    %origin:
    while (abs(t_grid(j)-t(i)) <= abs(t_grid(j)-t(next_i)))
        pnum(j) = i;
        j = j + 1;
        
        %No more output points:
        if (j > N_OUT) 
           break; 
        end
    end
    
    %No more output points?
    if (j > N_OUT) 
        break;
    end
end

%For every polynomial (i)
k = 1;
for i=1:N
    %Find limits:
    n_left = min([nl i-1]);
    n_right = min([nr N-i]);
    
    %Number of points in fit:
    num = 1 + n_left + n_right;
    
    %Create regressors and measurements:
    phi = zeros(m, num);
    z = zeros(num, 1);
    for j=-n_left:n_right
        %Regressor:
        for r=1:m
            phi(r, j + n_left + 1) = (t(i+j) - t(i))^r;
        end
        %Measurement:
        z(j + n_left + 1) = y(i+j);
    end

    %Find polynomial:
    theta = lse(z, phi');
    
    %Differentiate theta ld times:
    if (ld > 0)
        for l=1:ld
            for j=1:m
                theta(j) = j*theta(j+1);
            end
            theta(m+1) = 0;
        end
    end
    
    %For every output point associated with this polynomial:
    while (pnum(k) == i)
        
        %Create regressor for time t_grid(k):
        reg(1) = 1;
        for r=1:m
           reg(r+1, 1) = (t_grid(k)-t(i))^r;
        end
        
        %Set output to be the assiociated polynomial ld's derivative value at time t_grid(k).
        ret_y(k) = reg'*theta;
        
        %Next output point:
        k = k + 1;
       
        %Done?
        if (k > N_OUT)
            break;
        end
    end
    
    %Done?
    if (k > N_OUT)
        break;
    end
end

%Return output time grid:
ret_t = t_grid;
